import numpy as np
import pandas as pd
import pylogit as pl
from collections import OrderedDict
import tensorflow as tf
from sklearn.model_selection import train_test_split


NALT = 4
NCS = 1000
NVAR = 4
b = (2 * np.random.rand(NVAR, 1)) - 1
X = np.array([])
Y = np.array([])
G = np.array([])
F = np.array([])
for i in range(NCS):
    crtX = np.random.randn(NALT, NVAR)
    crtE = np.random.gumbel(0,1,(NALT, 1))
    # crtChosen = np.argmax(np.dot(crtX, b) + crtE)

    #---------------nonlinear1-----------------#
    crtX2 = np.hstack((crtX, crtX[:,0].reshape(crtX.shape[0],1)))
    crtX3 = np.zeros_like(crtX)
    for m in range(crtX3.shape[1]):
        crtX3[:,m] = crtX2[:,m] * crtX2[:,m+1]
        # # ---------------nonlinear2-----------------#
        # crtX3[:, m] = np.sqrt(np.abs(crtX2[:, m])) * crtX2[:, m + 1] ** 3

    # # ---------------nonlinear3-----------------#
    # crtX3 = np.zeros_like(crtX)
    # for m in range(crtX3.shape[1]):
    #     crtX3[:,m] = np.sin(crtX[:,m]) + np.square(crtX[:,m]) / np.exp(crtX[:,m])

    crtChosen = np.argmax(np.dot(crtX3, b) + crtE)
    crtY = np.zeros((NALT, 1))
    crtY[crtChosen] = 1
    crtG = np.full((NALT,1), i+1)
    if X.size == 0:
        X = crtX
        Y = crtY
        G = crtG
        F = crtX3
    else:
        X = np.vstack((X, crtX))
        Y = np.vstack((Y, crtY))
        G = np.vstack((G, crtG))
        F = np.vstack((F, crtX3))
# generate train and test data
cut_index = int(NCS*NALT*0.7)
X_train = X[:cut_index,:]
X_test = X[cut_index:,:]
Y_train = Y[:cut_index,:]
Y_test = Y[cut_index:,:]
F_train = F[:cut_index,:]
F_test = F[cut_index:,:]
G_train = G[:cut_index,:]
G_test = G[cut_index:,:]
alt = np.array([(i+1) for i in range(NALT)])
ALT = np.matlib.repmat(alt.reshape(NALT,1), NCS, 1)
ALT_train = ALT[:cut_index,:]
ALT_test = ALT[cut_index:,:]
varlist = ["X"+str(i+1) for i in range(NVAR)]
varlist.append('Y')
varlist.append('Group')
varlist.append('ALT')
data_train = pd.DataFrame(np.hstack((X_train, Y_train, G_train, ALT_train)), columns=varlist)
data_test = pd.DataFrame(np.hstack((X_test, Y_test, G_test, ALT_test)), columns=varlist)

#logit模型
spec = OrderedDict()
for i in range(NVAR):
    spec["X"+str(i+1)] = [alt.tolist()]
model_usingX = pl.create_choice_model(data = data_train,
                               alt_id_col="ALT",
                               obs_id_col="Group",
                               choice_col="Y",
                               specification=spec,
                               model_type = "MNL"
                               )
model_usingX.fit_mle(np.zeros(NVAR))
# model.print_summaries()
mnl_rst_usingX = model_usingX.params.tolist()

# logit for F
spec2 = OrderedDict()
data2_train = data_train.copy()
data2_train.iloc[:,:NVAR] = F_train
for i in range(NVAR):
    spec2["F"+str(i+1)] = [alt.tolist()]
model_usingF = pl.create_choice_model(data = data2_train,
                               alt_id_col="ALT",
                               obs_id_col="Group",
                               choice_col="Y",
                               specification=spec,
                               model_type = "MNL"
                               )
model_usingF.fit_mle(np.zeros(NVAR))
# model2.print_summaries()
mnl_rst_usingF = model_usingF.params.tolist()



#cnn3
REG_RATE = 0.04
TRAIN_STEPS = 5000
learning_rate = 0.01
cnnX = tf.placeholder(tf.float32, (None, NALT, 1, NVAR))
cnnY = tf.placeholder(tf.float32, (None, NALT))
# W = tf.get_variable('W1', [1,1,NVAR,1], initializer=tf.zeros_initializer())
regularizer = tf.contrib.layers.l2_regularizer(REG_RATE)
W1 = tf.get_variable('W1', [1,1,NVAR,20], initializer=tf.contrib.layers.xavier_initializer())
tf.add_to_collection("losses", regularizer(W1))
W2 = tf.get_variable('W2', [1,1,20,40], initializer=tf.contrib.layers.xavier_initializer())
tf.add_to_collection("losses", regularizer(W2))
W3 = tf.get_variable('W3',[1,1,40,1], initializer=tf.contrib.layers.xavier_initializer())
tf.add_to_collection("losses", regularizer(W3))
Z1 = tf.nn.relu(tf.nn.conv2d(cnnX, W1, (1,1,1,1),"VALID"))
Z2 = tf.nn.relu(tf.nn.conv2d(Z1, W2, (1,1,1,1), "VALID"))
Z3 = tf.nn.conv2d(Z2, W3, (1,1,1,1),"VALID")
Z4 = tf.contrib.layers.flatten(Z3)
SM = tf.nn.softmax(Z4)
cross_entropy_loss = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(labels=cnnY, logits=Z4))
loss = cross_entropy_loss + tf.add_n(tf.get_collection("losses"))
opt = tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(loss)
init = tf.global_variables_initializer()
X2_train = X_train.reshape(int(NCS*0.7), NALT, 1, NVAR)
Y2_train = Y_train.reshape(int(NCS*0.7), NALT)
with tf.Session() as sess:
    sess.run(init)
    for iter in range(TRAIN_STEPS):
        _, crt_w, crt_loss = sess.run([opt, W2, loss], feed_dict={cnnX:X2_train, cnnY:Y2_train})
        if iter % 500 == 0:
            print('After iteration %i, cost is %f' %(iter, crt_loss))
    cnn_rst = sess.run(W2)
    cnn_softmax_output_train = sess.run(SM, feed_dict={cnnX:X2_train, cnnY:Y2_train})
    X2_test = X_test.reshape(int(NCS * 0.3), NALT, 1, NVAR)
    Y2_test = Y_test.reshape(int(NCS * 0.3), NALT)
    cnn_softmax_output_test = sess.run(SM, feed_dict={cnnX:X2_test, cnnY:Y2_test})



#check the goddness of fit
chosen_prob_usingX_train = model_usingX.predict(data_train, choice_col="Y", return_long_probs=False)
print(chosen_prob_usingX_train)
ll_usingX_train = np.log(chosen_prob_usingX_train).sum()
r2_usingX_train = 1 - ll_usingX_train / (np.log(1/NALT) * NCS * 0.7)

chosen_prob_usingX_test = model_usingX.predict(data_test, choice_col="Y", return_long_probs=False)
ll_usingX_test = np.log(chosen_prob_usingX_test).sum()
r2_usingX_test = 1 - ll_usingX_test / (np.log(1/NALT) * NCS * 0.3)

chosen_prob_usingF_train = model_usingF.predict(data2_train, choice_col="Y", return_long_probs=False)
ll_usingF_train = np.log(chosen_prob_usingF_train).sum()
r2_usingF_train = 1 - ll_usingF_train / (np.log(1/NALT) * NCS * 0.7)

data2_test = data_test.copy()
data2_test.iloc[:,:NVAR] = F_test
chosen_prob_usingX_test = model_usingF.predict(data2_test, choice_col="Y", return_long_probs=False)
ll_usingF_test = np.log(chosen_prob_usingX_test).sum()
r2_usingF_test = 1 - ll_usingF_test / (np.log(1/NALT) * NCS * 0.3)

ll_cnn_train = Y_train.flatten().dot(np.log(cnn_softmax_output_train.flatten()))
r2_cnn_train = 1 - ll_cnn_train / (np.log(1/NALT) * NCS * 0.7)

ll_cnn_test = Y_test.flatten().dot(np.log(cnn_softmax_output_test.flatten()))
r2_cnn_test = 1 - ll_cnn_test / (np.log(1/NALT) * NCS * 0.3)

print("------------------------------\n\n For Training Data: ")
print(" Loglikehood of MNL-1, MNL-2 and CNN is %4.2f, %4.2f, %4.2f respectively" % (ll_usingX_train, ll_usingF_train, ll_cnn_train))
print(" R-square of MNL-1, MNL-2 and CNN is %4.4f, %4.4f, %4.4f respectively\n" % (r2_usingX_train, r2_usingF_train, r2_cnn_train))
print("\n\n For Test Data: ")
print(" Loglikehood of MNL-1, MNL-2 and CNN is %4.2f, %4.2f, %4.2f respectively" % (ll_usingX_test, ll_usingF_test, ll_cnn_test))
print(" R-square of MNL-1, MNL-2 and CNN is %4.4f, %4.4f, %4.4f respectively\n" % (r2_usingX_test, r2_usingF_test, r2_cnn_test))










