import numpy as np
import pandas as pd
import pylogit as pl
from collections import OrderedDict
import tensorflow as tf


NALT = 10           #number of alternatives
NCS = 1000          #number of choice situations (sample size)
NVAR = 10           #number of explanatory variables
b = (2 * np.random.rand(NVAR, 1)) - 1           #real coefficients
# generate choice data (X:explantory variables, Y: choice results; G:group)
X = np.array([])                                         
Y = np.array([])
G = np.array([])
for i in range(NCS):
    crtX = np.random.randn(NALT, NVAR)
    crtE = np.random.gumbel(0,1,(NALT, 1))
    crtChosen = np.argmax(np.dot(crtX, b) + crtE)
    crtY = np.zeros((NALT, 1))
    crtY[crtChosen] = 1
    crtG = np.full((NALT,1), i+1)
    if X.size == 0:
        X = crtX
        Y = crtY
        G = crtG
    else:
        X = np.vstack((X, crtX))
        Y = np.vstack((Y, crtY))
        G = np.vstack((G, crtG))
varlist = ["X"+str(i+1) for i in range(NVAR)]
varlist.append('Y')
varlist.append('Group')
varlist.append('ALT')
data = np.hstack((X, Y, G))
alt = np.array([(i+1) for i in range(NALT)])
ALT = np.matlib.repmat(alt.reshape(NALT,1), NCS, 1)
data = np.hstack((data, ALT))
data = pd.DataFrame(data, columns=varlist)
data.to_csv("./data.csv", index=False)
Beta = pd.DataFrame(b, index=["b"+str(i+1) for i in range(NVAR)], columns=None)
Beta.to_csv("./Beta.csv", index=True, header=False)

# specify and estimate MNL model using pylogit package
spec = OrderedDict()
for i in range(NVAR):
    spec["X"+str(i+1)] = [alt.tolist()]
print('\n\n--------------Starting MNL estimation--------------')
model = pl.create_choice_model(data = data,
                               alt_id_col="ALT",
                               obs_id_col="Group",
                               choice_col="Y",
                               specification=spec,
                               model_type = "MNL"
                               )
model.fit_mle(np.zeros(NVAR))
#model.print_summaries()
mnl_rst = model.params.tolist()


# specify and estimate cnn model using tensorflow
print('\n\n--------------Starting CNN estimation--------------')
TRAIN_STEPS = 1000
learning_rate = 0.03
cnnX = tf.placeholder(tf.float32, (None, NALT, 1, NVAR))
cnnY = tf.placeholder(tf.float32, (None, NALT))
W = tf.get_variable('W1', [1,1,NVAR,1], initializer=tf.zeros_initializer())
Z = tf.nn.conv2d(cnnX, W, (1,1,1,1),"VALID")
Z2 = tf.contrib.layers.flatten(Z)
loss = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(labels=cnnY, logits=Z2))
opt = tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(loss)
init = tf.global_variables_initializer()
X2 = X.reshape(NCS, NALT, 1, NVAR)
Y2 = Y.reshape(NCS, NALT)
with tf.Session() as sess:
    sess.run(init)
    for iter in range(TRAIN_STEPS):
        _, crt_w, crt_loss = sess.run([opt, W, loss], feed_dict={cnnX:X2, cnnY:Y2})
        if iter % 200 == 0:
            print('After iteration %i, cost is %f' %(iter, crt_loss))
    cnn_rst = sess.run(W)

# compare the coefficients derived from MNL and CNN
print('\n\n--------------Comparison of coefficients--------------')
compB = pd.DataFrame(
    data=np.hstack((b, np.array(mnl_rst).reshape(NVAR,1), cnn_rst.reshape(NVAR,1))),
    index=varlist[:NVAR], columns=["Real","MNL", "CNN"])
print(compB)







