import numpy as np
import pandas as pd
import pylogit as pl
from collections import OrderedDict
import tensorflow as tf
from numpy.matlib import *

def to_choose(P):
    P = P / np.sum(P)
    cumP = np.insert(np.ndarray.cumsum(P), 0 ,0)
    r = np.random.rand()
    temp = (cumP - r) > 0
    rst = np.where(temp)[0][0] - 1
    return rst


NALT = 6
NCS = 3000
NVAR = 10
b = (2 * np.random.rand(NVAR, 1)) - 1
bnest = np.random.rand(int(NALT/2), 1) / 3
bnest_temp = repmat(bnest.reshape(bnest.size,1),1,2)
bnest_temp = bnest_temp.reshape(NALT, 1)
X = np.array([])
Y = np.array([])
G = np.array([])
for i in range(NCS):
    crtX = np.random.randn(NALT, NVAR)
    crt_expV_divide_lambda = np.exp(np.dot(crtX, b).reshape(NALT,1) / bnest_temp)
    temp1 = np.sum(crt_expV_divide_lambda.reshape(int(NALT/2),2), axis=1, keepdims=True)
    temp2 = temp1 ** (bnest.reshape(bnest.size,1) - 1)
    temp2 = repmat(temp2.reshape(temp2.size,1),1,2).reshape(temp2.size*2,1)
    numerator = crt_expV_divide_lambda * temp2
    denominator = np.sum(temp1 ** bnest.reshape(bnest.size,1))
    crtP = numerator / denominator
    crtChosen = to_choose(crtP)
    crtY = np.zeros((NALT, 1))
    crtY[crtChosen] = 1
    crtG = np.full((NALT,1), i+1)
    if X.size == 0:
        X = crtX
        Y = crtY
        G = crtG
    else:
        X = np.vstack((X, crtX))
        Y = np.vstack((Y, crtY))
        G = np.vstack((G, crtG))
# generate train and test data
cut_index = int(NCS*NALT*0.7)
X_train = X[:cut_index,:]
X_test = X[cut_index:,:]
Y_train = Y[:cut_index,:]
Y_test = Y[cut_index:,:]
G_train = G[:cut_index,:]
G_test = G[cut_index:,:]
alt = np.array([(i+1) for i in range(NALT)])
ALT = repmat(alt.reshape(NALT,1), NCS, 1)
ALT_train = ALT[:cut_index,:]
ALT_test = ALT[cut_index:,:]
varlist = ["X"+str(i+1) for i in range(NVAR)]
varlist.append('Y')
varlist.append('Group')
varlist.append('ALT')
data_train = pd.DataFrame(np.hstack((X_train, Y_train, G_train, ALT_train)), columns=varlist)
data_test = pd.DataFrame(np.hstack((X_test, Y_test, G_test, ALT_test)), columns=varlist)
data_train.to_csv("./data.csv", index=False)

#nested logit model
spec = OrderedDict()
nest_membership = OrderedDict()
for nest_i in range(int(NALT/2)):
    nest_membership["Nest"+str(nest_i+1)] = [nest_i*2+1, nest_i*2+2]
for i in range(NVAR):
    spec["X"+str(i+1)] = [alt.tolist()]
model_nl = pl.create_choice_model(data = data_train,
                               alt_id_col="ALT",
                               obs_id_col="Group",
                               choice_col="Y",
                               specification=spec,
                               model_type = "Nested Logit",
                               nest_spec = nest_membership
                               )
model_nl.fit_mle(np.zeros(NVAR+ int(NALT/2)))
# model_nl.print_summaries()
nl_rst = np.array(model_nl.params.tolist())
# print(b)
# print(bnest)

#logit model
model_mnl = pl.create_choice_model(data = data_train,
                               alt_id_col="ALT",
                               obs_id_col="Group",
                               choice_col="Y",
                               specification=spec,
                               model_type = "MNL",
                               )
model_mnl.fit_mle(np.zeros(NVAR))
# model_mnl.print_summaries()
mnl_rst = np.array(model_mnl.params.tolist())

#compare the coefficient
nl_rst2 = nl_rst[int(NALT/2):]
comp_logit = pd.DataFrame(
    np.hstack((np.array(b).reshape(b.size,1), nl_rst2.reshape(nl_rst2.size,1), mnl_rst.reshape(mnl_rst.size,1))),
    columns=["Real","NL","MNL"],index=varlist[:NVAR]
)
print('\n\n--------------Comparison of coefficients--------------')
print(comp_logit)
print("\n\n")


#cnn
REG_RATE = 0.02
TRAIN_STEPS = 5000
learning_rate = 0.01
cnnX = tf.placeholder(tf.float32, (None, NALT, 1, NVAR))
cnnY = tf.placeholder(tf.float32, (None, NALT))
regularizer = tf.contrib.layers.l2_regularizer(REG_RATE)
W1 = tf.get_variable('W1', [1,1,NVAR,20], initializer=tf.contrib.layers.xavier_initializer())
tf.add_to_collection("losses", regularizer(W1))
W2 = tf.get_variable('W2', [1,1,20,40], initializer=tf.contrib.layers.xavier_initializer())
tf.add_to_collection("losses", regularizer(W2))
W3 = tf.get_variable('W3',[1,1,40,1], initializer=tf.contrib.layers.xavier_initializer())
tf.add_to_collection("losses", regularizer(W3))
Z1 = tf.nn.relu(tf.nn.conv2d(cnnX, W1, (1,1,1,1),"VALID"))
Z2 = tf.nn.relu(tf.nn.conv2d(Z1, W2, (1,1,1,1), "VALID"))
Z3 = tf.nn.conv2d(Z2, W3, (1,1,1,1),"VALID")
Z4 = tf.contrib.layers.flatten(Z3)
SM = tf.nn.softmax(Z4)
cross_entropy_loss = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(labels=cnnY, logits=Z4))
loss = cross_entropy_loss + tf.add_n(tf.get_collection("losses"))
opt = tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(loss)
init = tf.global_variables_initializer()
X2_train = X_train.reshape(int(NCS*0.7), NALT, 1, NVAR)
Y2_train = Y_train.reshape(int(NCS*0.7), NALT)
with tf.Session() as sess:
    sess.run(init)
    for iter in range(TRAIN_STEPS):
        _, crt_w, crt_loss = sess.run([opt, W2, loss], feed_dict={cnnX:X2_train, cnnY:Y2_train})
        if iter % 1000 == 0:
            print('After iteration %i, cost is %f' %(iter, crt_loss))
    cnn_rst = sess.run(W2)
    cnn_softmax_output_train = sess.run(SM, feed_dict={cnnX:X2_train, cnnY:Y2_train})
    X2_test = X_test.reshape(int(NCS * 0.3), NALT, 1, NVAR)
    Y2_test = Y_test.reshape(int(NCS * 0.3), NALT)
    cnn_softmax_output_test = sess.run(SM, feed_dict={cnnX:X2_test, cnnY:Y2_test})



#check the goddness of fit
chosen_prob_mnl_train = model_mnl.predict(data_train, choice_col="Y", return_long_probs=False)
ll_mnl_train = np.log(chosen_prob_mnl_train).sum()
r2_mnl_train = 1 - ll_mnl_train / (np.log(1/NALT) * NCS * 0.7)

chosen_prob_mnl_test = model_mnl.predict(data_test, choice_col="Y", return_long_probs=False)
ll_mnl_test = np.log(chosen_prob_mnl_test).sum()
r2_mnl_test = 1 - ll_mnl_test / (np.log(1/NALT) * NCS * 0.3)

chosen_prob_nl_train = model_nl.predict(data_train, choice_col="Y", return_long_probs=False)
ll_nl_train = np.log(chosen_prob_nl_train).sum()
r2_nl_train = 1 - ll_nl_train / (np.log(1/NALT) * NCS * 0.7)

chosen_prob_nl_test = model_nl.predict(data_test, choice_col="Y", return_long_probs=False)
ll_nl_test = np.log(chosen_prob_nl_test).sum()
r2_nl_test = 1 - ll_nl_test / (np.log(1/NALT) * NCS * 0.3)

ll_cnn_train = Y_train.flatten().dot(np.log(cnn_softmax_output_train.flatten()))
r2_cnn_train = 1 - ll_cnn_train / (np.log(1/NALT) * NCS * 0.7)

ll_cnn_test = Y_test.flatten().dot(np.log(cnn_softmax_output_test.flatten()))
r2_cnn_test = 1 - ll_cnn_test / (np.log(1/NALT) * NCS * 0.3)

print("------------------------------\n\n   For Training Data: ")
print(" Loglikehood of MNL, nested-logit and cnn is %4.2f, %4.2f, %4.2f respectively" % (ll_mnl_train, ll_nl_train, ll_cnn_train))
print(" R-square of MNL, nested-logit and cnn is %4.4f, %4.4f, %4.4f respectively\n" % (r2_mnl_train, r2_nl_train, r2_cnn_train))
print("\n\n For Test Data: ")
print(" Loglikehood of MNL, nested-logit and cnn is %4.2f, %4.2f, %4.2f respectively" % (ll_mnl_test, ll_nl_test, ll_cnn_test))
print(" R-square of MNL, nested-logit and cnn is %4.4f, %4.4f, %4.4f respectively\n" % (r2_mnl_test, r2_nl_test, r2_cnn_test))










